/*---------------------------------------------------------------------------
 * Copyright (C) 1999 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

package com.dalsemi.onewire.container;

// imports
import com.dalsemi.onewire.OneWireException;
import com.dalsemi.onewire.adapter.OneWireIOException;
import com.dalsemi.onewire.utils.*;
import com.dalsemi.onewire.adapter.*;
import com.dalsemi.onewire.container.OneWireContainer;

/**
  * Memory bank class for the DS1961S/DS2432.
  *
  *  @version 	0.00, 19 Dec 2000
  *  @author 	DS
  */

class MemoryBankSHAEE
   implements PagedMemoryBank
{

   //--------
   //--------Static Final Variables
   //--------

   /** Read Memory Command */
   public static final byte READ_MEMORY = ( byte ) 0xF0;

   /** Write Scratch Pad */
   public static final byte WRITE_SCRATCHPAD = ( byte ) 0x0F;

   /** Read Scratch Pad */
   public static final byte READ_SCRATCHPAD = ( byte ) 0xAA;

   /** Load First Secret */
   public static final byte LOAD_FIRST_SECRET = ( byte ) 0x5A;

   /** Copy Scratch Pad */
   public static final byte COPY_SCRATCHPAD = ( byte ) 0x55;

   /** Read Authenticate Page */
   public static final byte READ_AUTH_PAGE = ( byte ) 0xA5;

   /** Compute next Secret command */
   public static final byte NEXT_SECRET = ( byte ) 0x33;

   //--------
   //-------- Protected Variables for MemoryBank implementation 
   //--------

   /**
     * Check the status of the memory page.
     */
   protected boolean checked;

   /**
     * Size of memory bank in bytes
     */
   protected int size;

   /**
    * Memory bank descriptions
    */
   protected String bankDescription;

   /**
    * Memory bank usage flags
    */
   protected boolean generalPurposeMemory;

   /**
    * Flag if memory bank is read/write
    */
   protected boolean readWrite;

   /**
    * Flag if memory bank is write once (EPROM)
    */
   protected boolean writeOnce;

   /**
    * Flag if memory bank is read only
    */
   protected boolean readOnly;

   /**
    * Starting physical address in memory bank.  Needed for different
    * types of memory in the same logical memory bank.  This can be
    * used to seperate them into two virtual memory banks.  Example:
    * DS2406 status page has mixed EPROM and Volatile RAM.
    */
   protected int startPhysicalAddress;

   /**
    * Flag if read back verification is enabled in 'write()'.
    */
   protected boolean writeVerification;

  /**
    * Number of pages in memory bank
    */
   protected int numberPages;

   /**
    *  page length in memory bank
    */
   protected int pageLength;

   /**
    * Flag if reading a page in memory bank provides optional
    * extra information (counter, tamper protection, SHA-1...)
    */
   protected boolean extraInfo;

   /**
    * Extra information length in bytes
    */
   protected int extraInfoLength;

   /**
    * Max data length in page packet in memory bank
    */
   protected int maxPacketDataLength;

   /**
    * Flag if memory bank has page CRC.
    */
   protected boolean pageCRC;

   //--------
   //-------- Variables
   //--------

   /**
    * Reference to the OneWireContainer this bank resides on.
    */
   protected OneWireContainer33 ib;

   /**
    * block of 0xFF's used for faster read pre-fill of 1-Wire blocks
    */
   protected byte[] ffBlock;

   /**
    * block of 0x00's used for faster read pre-fill of 1-Wire blocks
    */
   protected byte[] zeroBlock;

   /**
    * Flag to indicate that speed needs to be set
    */
   protected boolean doSetSpeed;

   /**
    * Memory bank constructor.  Requires reference to the OneWireContainer
    * this memory bank resides on.
    */
   public MemoryBankSHAEE (OneWireContainer33 ibutton)
   {
      int i;

      synchronized (this)
      {
         // keep reference to ibutton where memory bank is
         ib = ibutton;

         // create the ffblock (usef for faster 0xFF fills)
         ffBlock = new byte [50];
         // create the zeroblock
         zeroBlock = new byte [6];

         for(i=0; i<50; i++)
            ffBlock[i] = ( byte ) 0xFF;
         for(i=0;i<6;i++)
            zeroBlock[i] = ( byte ) 0x00;
         
         // indicate speed has not been set
         doSetSpeed = true;
      }
   }
 
   //--------
   //-------- Memory Bank methods
   //--------

   /**
    * Query to see get a string description of the current memory bank.
    *
    * @return  String containing the memory bank description
    */
   public String getBankDescription ()
   {
      return bankDescription;
   }

   /**
    * Query to see if the current memory bank is general purpose
    * user memory.  If it is NOT then it is Memory-Mapped and writing
    * values to this memory will affect the behavior of the 1-Wire
    * device.
    *
    * @return  'true' if current memory bank is general purpose
    */
   public boolean isGeneralPurposeMemory ()
   {
      return generalPurposeMemory;
   }

  /**
    * Query to see if current memory bank is read/write.
    *
    * @return  'true' if current memory bank is read/write
    */
   public boolean isReadWrite ()
   {
      return readWrite;
   }

   /**
    * Query to see if current memory bank is write write once such
    * as with EPROM technology.
    *
    * @return  'true' if current memory bank can only be written once
    */
   public boolean isWriteOnce ()
   {
      return writeOnce;
   }

   /**
    * Query to see if current memory bank is read only.
    *
    * @return  'true' if current memory bank can only be read
    */
   public boolean isReadOnly ()
   {
      return readOnly;
   }

   /**
    * Query to see if current memory bank non-volatile.  Memory is
    * non-volatile if it retains its contents even when removed from
    * the 1-Wire network.
    *
    * @return  'true' if current memory bank non volatile.
    */
   public boolean isNonVolatile ()
   {
      return true;
   }

  /**
    * Query to see if current memory bank pages need the adapter to
    * have a 'ProgramPulse' in order to write to the memory.
    *
    * @return  'true' if writing to the current memory bank pages
    *                 requires a 'ProgramPulse'.
    */
   public boolean needsProgramPulse ()
   {
      return false;
   }

   /**
    * Query to see if current memory bank pages need the adapter to
    * have a 'PowerDelivery' feature in order to write to the memory.
    *
    * @return  'true' if writing to the current memory bank pages
    *                 requires 'PowerDelivery'.
    */
   public boolean needsPowerDelivery ()
   {
      return false;
   }

   /**
    * Query to get the starting physical address of this bank.  Physical
    * banks are sometimes sub-divided into logical banks due to changes
    * in attributes.
    *
    * @return  physical starting address of this logical bank.
    */
   public int getStartPhysicalAddress ()
   {
      return startPhysicalAddress;
   }

   /**
    * Query to get the memory bank size in bytes.
    *
    * @return  memory bank size in bytes.
    */
   public int getSize ()
   {
      return size;
   }

   /**
    * Set the write verification for the 'write()' method.
    *
    * @param  doReadVerf   true (default) verify write in 'write'
    *                      false, don't verify write (used on
    *                      Write-Once bit manipulation)
    */
   public void setWriteVerification (boolean doReadVerf)
   {
      writeVerification = doReadVerf;
   }

   /**
    * Query to get the number of pages in current memory bank.
    *
    * @return  number of pages in current memory bank
    */
   public int getNumberPages ()
   {
      return numberPages;
   }

   /**
    * Query to get  page length in bytes in current memory bank.
    *
    * @return   page length in bytes in current memory bank
    */
   public int getPageLength ()
   {
      return pageLength;
   }

   /**
    * Query to see if current memory bank pages can be read with
    * the contents being verified by a device generated CRC.
    * This is used to see if the 'ReadPageCRC()' can be used.
    *
    * @return  'true' if current memory bank can be read with self
    *          generated CRC.
    */
   public boolean hasPageAutoCRC ()
   {
      return pageCRC;
   }

   /**
    * Query to see if current memory bank pages when read deliver
    * extra information outside of the normal data space.  Examples
    * of this may be a redirection byte, counter, tamper protection
    * bytes, or SHA-1 result.  If this method returns true then the
    * methods 'ReadPagePacket()' and 'readPageCRC()' with 'extraInfo'
    * parameter can be used.
    *
    * @return  'true' if reading the current memory bank pages
    *                 provides extra information.
    *
    * @deprecated  As of 1-Wire API 0.01, replaced by {@link #hasExtraInfo()}
    */
   public boolean haveExtraInfo ()
   {
      return extraInfo;
   }

   /**
    * Checks to see if this memory bank's pages deliver extra 
    * information outside of the normal data space,  when read.  Examples
    * of this may be a redirection byte, counter, tamper protection
    * bytes, or SHA-1 result.  If this method returns true then the
    * methods with an 'extraInfo' parameter can be used:
    * {@link #readPage(int,boolean,byte[],int,byte[]) readPage},
    * {@link #readPageCRC(int,boolean,byte[],int,byte[]) readPageCRC}, and
    * {@link #readPagePacket(int,boolean,byte[],int,byte[]) readPagePacket}.
    *
    * @return  <CODE> true </CODE> if reading the this memory bank's 
    *                 pages provides extra information
    *
    * @see #readPage(int,boolean,byte[],int,byte[]) readPage(extra)
    * @see #readPageCRC(int,boolean,byte[],int,byte[]) readPageCRC(extra)
    * @see #readPagePacket(int,boolean,byte[],int,byte[]) readPagePacket(extra)
    * @since 1-Wire API 0.01
    */
   public boolean hasExtraInfo ()
   {
      return extraInfo;
   }

   /**
    * Query to get the length in bytes of extra information that
    * is read when read a page in the current memory bank.  See
    * 'hasExtraInfo()'.
    *
    * @return  number of bytes in Extra Information read when reading
    *          pages in the current memory bank.
    */
   public int getExtraInfoLength ()
   {
      return extraInfoLength;
   }

   /**
    * Query to get a string description of what is contained in
    * the Extra Informationed return when reading pages in the current
    * memory bank.  See 'hasExtraInfo()'.
    *
    * @return string describing extra information.
    */
   public String getExtraInfoDescription ()
   {
      return "The MAC for the SHA Engine";
   }

   /**
    * Query to get Maximum data page length in bytes for a packet
    * read or written in the current memory bank.  See the 'ReadPagePacket()'
    * and 'WritePagePacket()' methods.  This method is only usefull
    * if the current memory bank is general purpose memory.
    *
    * @return  max packet page length in bytes in current memory bank
    */
   public int getMaxPacketDataLength ()
   {
      return maxPacketDataLength;
   }

   //--------
   //-------- MemoryBank I/O methods
   //--------

   /**
    * Read  memory in the current bank with no CRC checking (device or
    * data). The resulting data from this API may or may not be what is on
    * the 1-Wire device.  It is recommends that the data contain some kind
    * of checking (CRC) like in the readPagePacket() method or have
    * the 1-Wire device provide the CRC as in readPageCRC().  readPageCRC()
    * however is not supported on all memory types, see 'hasPageAutoCRC()'.
    * If neither is an option then this method could be called more
    * then once to at least verify that the same thing is read consistantly.
    *
    * @param  startAddr     starting physical address
    * @param  readContinue  if 'true' then device read is continued without
    *                       re-selecting.  This can only be used if the new
    *                       read() continious where the last one led off
    *                       and it is inside a 'beginExclusive/endExclusive'
    *                       block.
    * @param  readBuf       byte array to place read data into
    * @param  offset        offset into readBuf to place data
    * @param  len           length in bytes to read
    *
    * @throws OneWireIOException
    * @throws OneWireException
    */
   public void read (int startAddr, boolean readContinue, byte[] readBuf,
                     int offset, int len)
      throws OneWireIOException, OneWireException
   {
      int i;

      // attempt to put device at max desired speed
      if (!readContinue && ib.adapterSet())
         checkSpeed();

      // check if read exceeds memory
      if ((startAddr + len) > (pageLength * numberPages))
         throw new OneWireException("Read exceeds memory bank end.");

      // see if need to access the device
      if (!readContinue)
      {

         // select the device
         if (!ib.adapter.select(ib.address))
            throw new OneWireIOException("Device select failed.");

         // build start reading memory block
         int    addr    = startAddr + startPhysicalAddress;
         byte[] raw_buf = new byte [3];

         raw_buf [0] = READ_MEMORY;
         raw_buf [1] = ( byte ) (addr & 0xFF);
         raw_buf [2] = ( byte ) (((addr & 0xFFFF) >>> 8) & 0xFF);
   
         // do the first block for command, address
         ib.adapter.dataBlock(raw_buf, 0, 3);
      }

      // pre-fill readBuf with 0xFF 
      int pgs   = len / pageLength;
      int extra = len % pageLength;

      for (i = 0; i < pgs; i++)
         System.arraycopy(ffBlock, 0, readBuf, offset + i * pageLength,
                          pageLength);
      System.arraycopy(ffBlock, 0, readBuf, offset + pgs * pageLength, extra);

      // send second block to read data, return result
      ib.adapter.dataBlock(readBuf, offset, len);
   }

   /**
    * Write  memory in the current bank.  It is recommended that
    * when writing  data that some structure in the data is created
    * to provide error free reading back with read().  Or the
    * method 'writePagePacket()' could be used which automatically
    * wraps the data in a length and CRC.
    *
    * When using on Write-Once devices care must be taken to write into
    * into empty space.  If write() is used to write over an unlocked
    * page on a Write-Once device it will fail.  If write verification
    * is turned off with the method 'setWriteVerification(false)' then
    * the result will be an 'AND' of the existing data and the new data.
    *
    * @param  startAddr     starting address
    * @param  writeBuf      byte array containing data to write
    * @param  offset        offset into writeBuf to get data
    * @param  len           length in bytes to write
    *
    * @throws OneWireIOException
    * @throws OneWireException
    */
   public void write (int startAddr, byte[] writeBuf, int offset, int len)
      throws OneWireIOException, OneWireException
   {
      int i, cnt=0, room_left;

      if(!checked)
         checked = ib.checkStatus();

      // return if nothing to do
      if (len == 0)
         return;

      // attempt to put device at speed
      checkSpeed();

      // check to see if secret is set
      if(!ib.isContainerSecretSet())
         throw new OneWireException("Secret is not set.");

      // check if write exceeds memory
      if ((startAddr + len) > size)
         throw new OneWireException("Write exceeds memory bank end");

      // check if trying to write read only bank
      if (isReadOnly())
         throw new OneWireException("Trying to write read-only memory bank");

      // loop while still have pages to write
      int    startx     = 0, nextx = 0;   // (start and next index into writeBuf)
      byte[] raw_buf    = new byte [8];
      byte[] extra_buf  = new byte [getExtraInfoLength()];
      byte[] send_block = new byte [64];
      byte[] memory     = new byte [32];
      int    abs_addr   = startPhysicalAddress + startAddr;
      int    pl         = 8;

      read((startAddr/pageLength)*pageLength,false,memory,0,pageLength);

      do
      {
         // calculate room left in current page
         room_left = pl - ((abs_addr + startx) % pl);

         // check if block left will cross end of page
         if ((len - startx) > room_left)
            nextx = startx + room_left;
         else
            nextx = len;

         if(startAddr >= pageLength)
            System.arraycopy(memory,(((startx+startAddr)/8)*8)-32,raw_buf,0,8);
         else
            System.arraycopy(memory,(((startx+startAddr)/8)*8),raw_buf,0,8);

         if((nextx-startx) == 8)
            System.arraycopy(writeBuf,offset+startx,raw_buf,0,8);
         else
            if(((startAddr+nextx)%8) == 0)
               System.arraycopy(writeBuf,offset+startx,raw_buf,((startAddr+startx)%8),8-((startAddr+startx)%8));
            else
               System.arraycopy(writeBuf,offset+startx,raw_buf,((startAddr+startx)%8),
                               ((startAddr+nextx)%8)-((startAddr+startx)%8));

         // write the page of data to scratchpad
         if(!writeSpad(abs_addr+startx+room_left-8,raw_buf,0,8))
            throw new OneWireIOException("Invalid CRC16 in write");

         // Copy data from scratchpad into memory
         copySpad(abs_addr+startx+room_left-8,extra_buf,memory);

         if(startAddr >= pageLength)
            System.arraycopy(raw_buf,0,memory,(((startx+startAddr)/8)*8)-32,8);
         else
            System.arraycopy(raw_buf,0,memory,(((startx+startAddr)/8)*8),8);

         // point to next index
         startx = nextx;
      }
      while (nextx < len);
   }

   //--------
   //-------- PagedMemoryBank I/O methods
   //--------

   /**
    * Read  page in the current bank with no
    * CRC checking (device or data). The resulting data from this API
    * may or may not be what is on the 1-Wire device.  It is recommends
    * that the data contain some kind of checking (CRC) like in the
    * readPagePacket() method or have the 1-Wire device provide the
    * CRC as in readPageCRC().  readPageCRC() however is not
    * supported on all memory types, see 'hasPageAutoCRC()'.
    * If neither is an option then this method could be called more
    * then once to at least verify that the same thing is read consistantly.
    *
    * @param  page          page number to read packet from
    * @param  readContinue  if 'true' then device read is continued without
    *                       re-selecting.  This can only be used if the new
    *                       readPage() continious where the last one
    *                       led off and it is inside a
    *                       'beginExclusive/endExclusive' block.
    * @param  readBuf       byte array to place read data into
    * @param  offset        offset into readBuf to place data
    *
    * @throws OneWireIOException
    * @throws OneWireException
    */
   public void readPage (int page, boolean readContinue, byte[] readBuf,
                         int offset)
      throws OneWireIOException, OneWireException
   {
      read(page * pageLength, readContinue, readBuf, offset, pageLength);
   }

   /**
    * Read  page with extra information in the current bank with no
    * CRC checking (device or data). The resulting data from this API
    * may or may not be what is on the 1-Wire device.  It is recommends
    * that the data contain some kind of checking (CRC) like in the
    * readPagePacket() method or have the 1-Wire device provide the
    * CRC as in readPageCRC().  readPageCRC() however is not
    * supported on all memory types, see 'hasPageAutoCRC()'.
    * If neither is an option then this method could be called more
    * then once to at least verify that the same thing is read consistantly.
    * See the method 'hasExtraInfo()' for a description of the optional
    * extra information some devices have.
    *
    * @param  page          page number to read packet from
    * @param  readContinue  if 'true' then device read is continued without
    *                       re-selecting.  This can only be used if the new
    *                       readPage() continious where the last one
    *                       led off and it is inside a
    *                       'beginExclusive/endExclusive' block.
    * @param  readBuf       byte array to place read data into
    * @param  offset        offset into readBuf to place data
    * @param  extraInfo     byte array to put extra info read into
    *
    * @throws OneWireIOException
    * @throws OneWireException
    */
   public void readPage (int page, boolean readContinue, byte[] readBuf,
                         int offset, byte[] extraInfo)
      throws OneWireIOException, OneWireException
   {
      byte[] pg  = new byte [32];
      int i;

      if(!checked)
         checked = ib.checkStatus();

      if(!hasPageAutoCRC())
         throw new OneWireException("This memory bank doesn't have crc capabilities.");

      // attempt to put device at max desired speed
      if (!readContinue)
         checkSpeed();

      // build start reading memory block
      int    addr    = page * pageLength + startPhysicalAddress;

      if(!ReadAuthPage(addr,pg,extraInfo))
         throw new OneWireException("Read didn't work.");

      System.arraycopy(pg,0,readBuf,offset,32);
   }

   /**
    * Read a Universal Data Packet.
    *
    * The Universal Data Packet always starts on page boundaries but
    * can end anywhere in the page.  The structure specifies the length of
    * data bytes not including the length byte and the CRC16 bytes.
    * There is one length byte. The CRC16 is first initialized to
    * the page number.  This provides a check to verify the page that
    * was intended is being read.  The CRC16 is then calculated over
    * the length and data bytes.  The CRC16 is then inverted and stored
    * low byte first followed by the high byte.  This is structure is
    * used by this method to verify the data but is not returned, only
    * the data payload is returned.
    *
    * @param  page          page number to read packet from
    * @param  readContinue  if 'true' then device read is continued without
    *                       re-selecting.  This can only be used if the new
    *                       readPagePacket() continious where the last one
    *                       stopped and it is inside a
    *                       'beginExclusive/endExclusive' block.
    * @param  readBuf       byte array to put data read. Must have at least
    *                       'getMaxPacketDataLength()' elements.
    * @param  offset        offset into readBuf to place data
    *
    * @return  number of data bytes read from the device and written to
    *          readBuf at the offset.
    *
    * @throws OneWireIOException
    * @throws OneWireException
    */
   public int readPagePacket (int page, boolean readContinue, byte[] readBuf,
                              int offset)
      throws OneWireIOException, OneWireException
   {
      byte[] raw_buf = new byte [pageLength];

      // read the  page
      readPage(page, readContinue, raw_buf, 0);
      
      // check if length is realistic
      if ((raw_buf [0] & 0x00FF) > maxPacketDataLength)
         throw new OneWireIOException("Invalid length in packet.");

      // verify the CRC is correct
      if (CRC16.compute(raw_buf, 0, raw_buf [0] + 3, page) == 0x0000B001)
      {

         // extract the data out of the packet
         System.arraycopy(raw_buf, 1, readBuf, offset, raw_buf [0]);

         // return the length
         return raw_buf [0];
      }
      else
         throw new OneWireIOException("Invalid CRC16 in packet read.");
   }

   /**
    * Read a Universal Data Packet and extra information.  See the
    * method 'readPagePacket()' for a description of the packet structure.
    * See the method 'hasExtraInfo()' for a description of the optional
    * extra information some devices have.
    *
    * @param  page          page number to read packet from
    * @param  readContinue  if 'true' then device read is continued without
    *                       re-selecting.  This can only be used if the new
    *                       readPagePacket() continious where the last one
    *                       stopped and it is inside a
    *                       'beginExclusive/endExclusive' block.
    * @param  readBuf       byte array to put data read. Must have at least
    *                       'getMaxPacketDataLength()' elements.
    * @param  offset        offset into readBuf to place data
    * @param  extraInfo     byte array to put extra info read into
    *
    * @return  number of data bytes read from the device and written to
    *          readBuf at the offset.
    *
    * @throws OneWireIOException
    * @throws OneWireException
    */
   public int readPagePacket (int page, boolean readContinue, byte[] readBuf,
                              int offset, byte[] extraInfo)
      throws OneWireIOException, OneWireException
   {
      byte[] raw_buf = new byte [pageLength];
      int addr;

      if(!checked)
         checked = ib.checkStatus();

      if(!hasPageAutoCRC())
         throw new OneWireException("This memory bank page doesn't have CRC capabilites.");

      addr = startPhysicalAddress + page*pageLength;

      // read the  page
      ReadAuthPage(addr, raw_buf, extraInfo);
      

      // check if length is realistic
      if ((raw_buf [0] & 0x00FF) > maxPacketDataLength)
         throw new OneWireIOException("Invalid length in packet.");

      // verify the CRC is correct
      if (CRC16.compute(raw_buf, 0, raw_buf [0] + 3, page) == 0x0000B001)
      {

         // extract the data out of the packet
         System.arraycopy(raw_buf, 1, readBuf, offset, raw_buf [0]);

         // return the length
         return raw_buf [0];
      }
      else
         throw new OneWireIOException("Invalid CRC16 in packet read.");
   }

   /**
    * Write a Universal Data Packet.  See the method 'readPagePacket()'
    * for a description of the packet structure.
    *
    * @param  page          page number to write packet to
    * @param  writeBuf      data byte array to write
    * @param  offset        offset into writeBuf where data to write is
    * @param  len           number of bytes to write
    *
    * @throws OneWireIOException
    * @throws OneWireException
    */
   public void writePagePacket (int page, byte[] writeBuf, int offset,
                                int len)
      throws OneWireIOException, OneWireException
   {

      // make sure length does not exceed max
      if (len > maxPacketDataLength)
         throw new OneWireIOException(
            "Length of packet requested exceeds page size.");

      // see if this bank is general read/write
      if (!generalPurposeMemory)
         throw new OneWireException(
            "Current bank is not general purpose memory.");

      // construct the packet to write
      byte[] raw_buf = new byte [len + 3];

      raw_buf [0] = ( byte ) len;

      System.arraycopy(writeBuf, 0, raw_buf, 1, len);

      int crc = CRC16.compute(raw_buf, 0, len + 1, page);

      raw_buf [len + 1] = ( byte ) (~crc & 0xFF);
      raw_buf [len + 2] = ( byte ) (((~crc & 0xFFFF) >>> 8) & 0xFF);

      // write the packet, return result
      write(page * pageLength, raw_buf, 0, len + 3);
   }

   /**
    * Read a complete memory page with CRC verification provided by the
    * device.  Not supported by all devices.  See the method
    * 'hasPageAutoCRC()'.
    *
    * @param  page          page number to read
    * @param  readContinue  if 'true' then device read is continued without
    *                       re-selecting.  This can only be used if the new
    *                       readPagePacket() continious where the last one
    *                       stopped and it is inside a
    *                       'beginExclusive/endExclusive' block.
    * @param  readBuf       byte array to put data read. Must have at least
    *                       'getMaxPacketDataLength()' elements.
    * @param  offset        offset into readBuf to place data
    *
    * @throws OneWireIOException
    * @throws OneWireException
    */
   public void readPageCRC (int page, boolean readContinue, byte[] readBuf,
                            int offset)
      throws OneWireIOException, OneWireException
   {
      int addr,i;
      byte[] extra = new byte [20];
      byte[] pg  = new byte [32];

      if(!checked)
         checked = ib.checkStatus();

      if(!hasPageAutoCRC())
         throw new OneWireException("This memory bank doesn't have CRC capabilites.");

      addr = startPhysicalAddress + page*pageLength;

      if(!ReadAuthPage(addr,pg,extra))
         throw new OneWireException("Read didn't work.");

      System.arraycopy(pg,0,readBuf,offset,pageLength);
   }

   /**
    * Read a complete memory page with CRC verification provided by the
    * device with extra information.  Not supported by all devices.
    * See the method 'hasPageAutoCRC()'.
    * See the method 'hasExtraInfo()' for a description of the optional
    * extra information.
    *
    * @param  page          page number to read
    * @param  readContinue  if 'true' then device read is continued without
    *                       re-selecting.  This can only be used if the new
    *                       readPagePacket() continious where the last one
    *                       stopped and it is inside a
    *                       'beginExclusive/endExclusive' block.
    * @param  readBuf       byte array to put data read. Must have at least
    *                       'getMaxPacketDataLength()' elements.
    * @param  offset        offset into readBuf to place data
    * @param  extraInfo     byte array to put extra info read into
    *
    * @throws OneWireIOException
    * @throws OneWireException
    */
   public void readPageCRC (int page, boolean readContinue, byte[] readBuf,
                            int offset, byte[] extraInfo)
      throws OneWireIOException, OneWireException
   {
      int addr,i;
      byte[] pg  = new byte [32];

      if(!checked)
         checked = ib.checkStatus();

      if(!hasPageAutoCRC())
         throw new OneWireException("This memory bank doesn't have CRC capabilities.");

      addr = startPhysicalAddress + page*pageLength;

      if(!ReadAuthPage(addr,pg,extraInfo))
         throw new OneWireException("Read didn't work.");

      System.arraycopy(pg,0,readBuf,offset,pageLength);
   }

   // ------------------------
   // Setting status
   // ------------------------
   
   /**
    * Write protect the memory bank.
    */
   public void writeprotect()
   {
      readOnly = true;
      readWrite = false;
   }

   /**
    * Sets the EPROM mode for this page.
    */
   public void setEPROM()
   {
      writeOnce = true;
   }


   // ------------------------
   // Extras
   // ------------------------

   /**
    * Check the device speed if has not been done before or if
    * an error was detected.
    *
    * @throws OneWireIOException
    * @throws OneWireException
    */
   public void checkSpeed ()
      throws OneWireIOException, OneWireException
   {

      // only check the speed 
      if (doSetSpeed)
      {

         // attempt to set the correct speed and verify device present
         ib.doSpeed();

         // no execptions so clear flag 
         doSetSpeed = false;
      }

   }

   /**
    * Write to the Scratch Pad, which is a max of 8 bytes.
    *
    * @param  addr          the address to write the data to
    * @param  out_buf       byte array to write into scratch pad
    * @param  offset        offset into out_buf to write the data
    * @param  len           length of the write data
    *
    * @throws OneWireIOException
    * @throws OneWireException
    */
   public boolean writeSpad(int addr, byte[] out_buf, int offset, int len)
      throws OneWireIOException, OneWireException
   {
      byte[] send_block = new byte [64];
      int i, cnt;
      
      cnt = 0;

      // access the device 
      if (ib.adapter.select(ib.address))
      { 
         send_block [cnt++] = WRITE_SCRATCHPAD;
         send_block [cnt++] = ( byte ) (addr & 0xFF);
         send_block [cnt++] = ( byte ) (((addr & 0xFFFF) >>> 8) & 0xFF);

         System.arraycopy(out_buf,offset,send_block,cnt,len);
         cnt = cnt + len;

         send_block [cnt++] = ( byte ) 0xFF;
         send_block [cnt++] = ( byte ) 0xFF;

         ib.adapter.dataBlock(send_block, 0, cnt);

         // verify the CRC is correct
         if (CRC16.compute(send_block, 0, cnt) != 0x0000B001)
            throw new OneWireIOException("Invalid CRC16 in Writing Scratch Pad");
      }
      else
         throw new OneWireIOException("Device select failed.");    

      return true;   
   }

   /**
    * Copy what is in the Sratch Pad, which is a max of 8 bytes to a certain
    * address in memory.
    *
    * @param  addr          the address to copy the data to
    * @param  extra_buf     byte array to write the calculated MAC
    * @param  memory        byte array for memory of the page to be copied to
    *
    * @throws OneWireIOException
    * @throws OneWireException
    */
   public void copySpad(int addr, byte[] extra_buf, byte[] memory)
      throws OneWireIOException, OneWireException
   {
      short send_cnt=0;
      int tmpadd=0;
      byte[] send_block = new byte [80];
      byte[] scratch    = new byte [11];
      byte[] secret     = new byte [8];
      byte[] sn         = new byte [8];
      byte es=0,test=0;
      int i;
      byte[] MT         = new byte [64];
      int A=0,B=0,C=0,D=0,E=0;
      int Temp;

      // check to see if secret is set
      if(!ib.isContainerSecretSet())
         throw new OneWireException("Secret is not set.");

      sn = ib.getAddress();

      ib.getSecret(secret);

      System.arraycopy(secret,0,MT,0,4);

      System.arraycopy(memory,0,MT,4,32);

      if(readSpad(scratch))
         System.arraycopy(scratch,0,MT,32,8);
      es = scratch[8];
      tmpadd = ((scratch[10]&0x00FF)<<8) | (scratch[9]&0x00FF);

      MT[40] = ( byte ) ((tmpadd & 0x00f0) >>> 5);

      System.arraycopy(sn,0,MT,41,7);

      System.arraycopy(secret,4,MT,48,4);

      System.arraycopy(ffBlock,0,MT,52,3);

      MT[55] = ( byte ) 0x80;

      System.arraycopy(zeroBlock,0,MT,56,6);

      MT[62] = ( byte ) 0x01;
      MT[63] = ( byte ) 0xB8;

      int[] AtoE = new int [5]; 
      ib.ComputeSHA(MT,AtoE);
      A = AtoE[0];
      B = AtoE[1];
      C = AtoE[2];
      D = AtoE[3];
      E = AtoE[4];

      // access the device 
      if (ib.adapter.select(ib.address))
      { 
         // Copy command
         send_block[send_cnt++] = COPY_SCRATCHPAD;
      
         // address 1
         send_block[send_cnt++] = ( byte ) (tmpadd & 0xFF);
         // address 2
         send_block[send_cnt++] = ( byte ) (((tmpadd & 0xFFFF) >>> 8) & 0xFF);

         // ending address with data status
         send_block[send_cnt++] = es;

         ib.adapter.dataBlock(send_block, 0, send_cnt);

         send_cnt = 0;
         
         // sending MAC
         Temp = E;
         for(i=0;i<4;i++)
         {
            send_block[send_cnt++] = ( byte ) (Temp & 0x000000FF);
            Temp >>= 8;
         }

         Temp = D;
         for(i=0;i<4;i++)
         {
            send_block[send_cnt++] = ( byte ) (Temp & 0x000000FF);
            Temp >>= 8;
         }

         Temp = C;
         for(i=0;i<4;i++)
         {
            send_block[send_cnt++] = ( byte ) (Temp & 0x000000FF);
            Temp >>= 8;
         }

         Temp = B;
         for(i=0;i<4;i++)
         {
            send_block[send_cnt++] = ( byte ) (Temp & 0x000000FF);
            Temp >>= 8;
         }

         Temp = A;
         for(i=0;i<4;i++)
         {
            send_block[send_cnt++] = ( byte ) (Temp & 0x000000FF);
            Temp >>= 8;
         }

         System.arraycopy(send_block,0,extra_buf,0,20);

         try
         {
            Thread.sleep(4);
         }
         catch (InterruptedException e){}

         ib.adapter.dataBlock(send_block,0,send_cnt);

         try
         {
            Thread.sleep(30);
         }
         catch (InterruptedException e){}     

         test = ( byte ) ib.adapter.getByte();

         if((test != ( byte ) 0xAA) && (test != ( byte ) 0x55))
         {
            if(test == ( byte ) 0xFF)
               throw new OneWireException("That area of memory is write-protected.");
            else if(test == ( byte )0x00)
               throw new OneWireIOException("Error due to not matching MAC.");
         }

      }
      else
         throw new OneWireIOException("Device select failed.");
   }         

   /**
    * Read from the Scratch Pad, which is a max of 8 bytes.
    *
    * @param  addr          the address to read the data
    * @param  es            offset byte read from scratchpad
    * @param  scratch       data buffer read from scratchpad
    *
    * @throws OneWireIOException
    * @throws OneWireException
    */
   public boolean readSpad(byte[] scratch)
      throws OneWireIOException, OneWireException
   {
      int send_cnt=0;
      byte[] send_block = new byte [50];
      int i;

      // access the device 
      if (ib.adapter.select(ib.address))
      {
         // read scratchpad command
         send_block[send_cnt++] = READ_SCRATCHPAD;
         // now add the read bytes for data bytes and crc16
         System.arraycopy(ffBlock,0,send_block,send_cnt,13);
         send_cnt = send_cnt + 13;

         // now send the block
         ib.adapter.dataBlock(send_block,0,send_cnt);
         
         // copy data to return buffers
         scratch[9] = send_block[1];
         scratch[10] = send_block[2];

         scratch[8] = send_block[3];

         // verify CRC16 is correct
         if (CRC16.compute(send_block, 0, send_cnt) == 0x0000B001)
         {
            System.arraycopy(send_block,4,scratch,0,8);
            // success
            return true;
         }
         else
            throw new OneWireException("Error due to CRC.");
       
      }
      else
         throw new OneWireIOException("Device select failed.");
   }

   /**
    *  Load First Secret for the DS2432.
    *
    * @param  addr          the address to write the data to 
    * @param  load          the first secret data
    *
    * @throws OneWireIOException
    * @throws OneWireException
    */
   public boolean LoadFirSecret(int addr, byte[] load)
      throws OneWireIOException, OneWireException
   {
      int tmpadd=0;
      byte   test=0;
      byte[] data = new byte[11];
      byte[] send_block = new byte[32];
      int send_cnt = 0;

      if(ib.isSecretWriteProtected())
         throw new OneWireException("Error due to secret is write-protected.");

      if(writeSpad(addr + startPhysicalAddress,load,0,8) &&
         readSpad(data))
      {
         tmpadd = ((data[10]&0x00FF)<<8) | (data[9]&0x00FF);
         // access the device 
         if (ib.adapter.select(ib.address))
         {         
            send_block [send_cnt++] = LOAD_FIRST_SECRET;
           
            // address 1
            send_block[send_cnt++] = data[9];
            // address 2
            send_block[send_cnt++] = data[10];
            // ending address with data status
            send_block [send_cnt++] = data[8];

            // now send the block
            ib.adapter.dataBlock(send_block,0,send_cnt);  

            try
            {
               Thread.sleep(10);
            }
            catch (InterruptedException e){}

            test = ( byte ) ib.adapter.getByte();

            if((test == (byte) 0xAA) || (test == (byte) 0x55))
            {
               ib.setContainerSecret(load);
               return true;
            }
            else
               throw new OneWireException("Error due to invalid load.");
         }
         else
            throw new OneWireIOException("Device select failed.");
      }

      return false;
   }          
 
   /**
    * Computes the next secret.
    *
    * @param  addr          the address to copy the data to
    * @param  extra_buf     byte array to write the calculated MAC
    *
    * @throws OneWireIOException
    * @throws OneWireException
    */
   public void nextSecret(int addr, byte[] extra_buf, byte[] partialsecret)
      throws OneWireIOException, OneWireException
   {
      int send_cnt=0;
      int tmpadd=0;
      byte[] send_block = new byte [64];
      byte[] memory     = new byte [32];
      byte[] secret     = new byte [8];
      byte[] scratch    = new byte [8];
      byte es=0,test=0;
      int i;
      byte[] MT     = new byte [64];
      int A=0,B=0,C=0,D=0,E=0;
      int Temp;

      // check to see if secret is set
      if(!ib.isContainerSecretSet())
         throw new OneWireException("Secret is not set.");

      read((addr/32)*32,false,memory,0,32);

      if(!writeSpad(addr,partialsecret,0,8))
         throw new OneWireIOException("Error writing scratch pad.");

      ib.getSecret(secret);

      System.arraycopy(secret,0,MT,0,4);

      System.arraycopy(memory,0,MT,4,32);

      System.arraycopy(ffBlock,0,MT,36,4);

      MT[40] = ( byte ) (partialsecret[0] & ( byte ) 0x3F);

      System.arraycopy(partialsecret,1,MT,41,7);

      System.arraycopy(secret,4,MT,48,4);

      System.arraycopy(ffBlock,0,MT,52,3);

      MT[55] = ( byte ) 0x80;

      System.arraycopy(zeroBlock,0,MT,56,6);

      MT[62] = ( byte ) 0x01;
      MT[63] = ( byte ) 0xB8;

      int[] AtoE = new int [5]; 
      ib.ComputeSHA(MT,AtoE);
      A = AtoE[0];
      B = AtoE[1];
      C = AtoE[2];
      D = AtoE[3];
      E = AtoE[4];

      Temp = E;
      for(i=0;i<4;i++)
      {
         secret[i] = ( byte ) (Temp & 0x000000FF);
         Temp >>= 8;
         extra_buf [send_cnt++] = secret [i];
      }

      Temp = D;
      for(i=0;i<4;i++)
      {
         secret[i+4] = ( byte ) (Temp & 0x000000FF);
         Temp >>= 8;
         extra_buf [send_cnt++] = secret [i];
      }

      Temp = C;
      for(i=0;i<4;i++)
      {
            extra_buf [send_cnt++] = ( byte ) (Temp & 0x000000FF);
            Temp >>= 8;            
      }

      Temp = B;
      for(i=0;i<4;i++)
      {
         extra_buf [send_cnt++] = ( byte ) (Temp & 0x000000FF);
         Temp >>= 8;
      }

      Temp = A;
      for(i=0;i<4;i++)
      {
         extra_buf [send_cnt++] = ( byte ) (Temp & 0x000000FF);
         Temp >>= 8;
      }

      send_cnt = 0;

      // access the device 
      if (ib.adapter.select(ib.address))
      { 
         // Next Secret command
         send_block[send_cnt++] = NEXT_SECRET;
      
         // address 1
         send_block[send_cnt++] = ( byte ) (addr & 0xFF);
         // address 2
         send_block[send_cnt++] = ( byte ) (((addr & 0xFFFF) >>> 8) & 0xFF);

         // now send the block
         ib.adapter.dataBlock(send_block,0,send_cnt); 

         try
         {
            Thread.sleep(24);
         }
         catch (InterruptedException e){}

         if(readSpad(scratch))
         {
            for(i=0;i<8;i++)
               if(scratch[i] != ( byte ) 0xAA)  
               {
                  hexPrint(scratch,0,8);
                  throw new OneWireIOException("Next secret not calculated.");
               }
         

            ib.setContainerSecret(secret);
         }
         else 
            throw new OneWireIOException("Error reading scratch pad.");
      }
      else
         throw new OneWireIOException("Device select failed.");         
   }   

   /**
    *  Reads authenticated page.
    *
    * @param addr          address of the data to be read along with Physical address
    * @param data          the data read from the address
    * @param extra_info    the MAC calculated for this function
    *
    * @throws OneWireIOException
    * @throws OneWireException
    */
   public boolean ReadAuthPage(int addr, byte[] data, byte[] extra_info)
      throws OneWireException, OneWireIOException
   {
      int send_cnt = 0;
      byte[] send_block = new byte [64];
      byte[] challenge  = new byte [8];
      int i;
   
      ib.getChallenge(challenge);

      if(!writeSpad(addr,challenge,0,8))
         throw new OneWireIOException("Error writing scratch pad.");
 
      // access the device 
      if (ib.adapter.select(ib.address))
      {
         // Read Authenticated Command
         send_block[send_cnt++] = READ_AUTH_PAGE;

         // address 1
         send_block[send_cnt++] = ( byte ) (addr & 0xFF);
         // address 2
         send_block[send_cnt++] = ( byte ) (((addr & 0xFFFF) >>> 8) & 0xFF);

         // data + FF byte 
         System.arraycopy(ffBlock,0,send_block,send_cnt,35);
         send_cnt = send_cnt + 35;

         // now send the block
         ib.adapter.dataBlock(send_block,0,send_cnt);
      
         // verify CRC16 is correct
         if (CRC16.compute(send_block,0,send_cnt,0) != 0x0000B001)
            throw new OneWireException("First CRC didn't pass.");

         System.arraycopy(send_block,3,data,0,32);

         send_cnt = 22;
         System.arraycopy(ffBlock,0,send_block,0,22);

         ib.adapter.dataBlock(send_block,0,send_cnt);

         // verify CRC16 is correct
         if (CRC16.compute(send_block, 0, send_cnt,0) != 0x0000B001)
            throw new OneWireException("Second CRC didn't pass.");
         else
            System.arraycopy(send_block,0,extra_info,0,20);
      }
      else
         throw new OneWireIOException("Device select failed.");

      return true;
   }

   /**
    * Print an array of bytes in hex to standard out.
    *
    * @param  dataBuf data to print
    * @param  offset  offset into dataBuf to start
    * @param  len     length of data to print
    */
   public static void hexPrint (byte[] dataBuf, int offset, int len)
   {
      for (int i = 0; i < len; i++)
      {
         if ((dataBuf [i + offset] & 0x000000FF) < 0x00000010)
         {
            System.out.print("0");
            System.out.print(Integer.toHexString(( int ) dataBuf [i + offset]
                                                 & 0x0000000F).toUpperCase());
         }
         else
            System.out.print(Integer.toHexString(( int ) dataBuf [i + offset]
                                                 & 0x000000FF).toUpperCase());
      }
   }

}
