
/*---------------------------------------------------------------------------
 * Copyright (C) 1999,2000 Dallas Semiconductor Corporation, All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Dallas Semiconductor
 * shall not be used except as stated in the Dallas Semiconductor
 * Branding Policy.
 *---------------------------------------------------------------------------
 */

// OneWireContainer16.java
package com.dalsemi.onewire.container;

import java.io.*;
import java.lang.*;
import java.util.*;
import com.dalsemi.onewire.*;
import com.dalsemi.onewire.utils.*;
import com.dalsemi.onewire.adapter.*;


/**
 * <P> 1-Wire container for Java-powered iButton, DS195X.  This container
 * encapsulates the functionality of the iButton family  type 
 * <B>16</B> and <B>96</B> (hex)</P> 
 *
 * <H3> Features </H3> 
 * <UL>
 *   <LI> Java Card&#8482 2.0-compliant
 *   <LI> True 32-bit Java integers for straightforward computation 
 *   <LI> Automatic garbage collection for efficient reuse of memory space 
 *   <LI> Resizable scratchpad optimizes memory usage and allows large atomic transactions 
 *   <LI> Dynamic applet capability for post-issuance updates 
 *   <LI> up to 64-kbyte ROM for Java VM and operating system 
 *   <LI> 6-kbyte NV SRAMwrites in 100 nanosecondssupports multiple applications 
 *   <LI> Java-accessible True Time Clock time-stamps transactions 
 *   <LI> Java-accessible random number generator for cryptographic keys 
 *   <LI> JavaCardX.Crypto, SHA-1, RSA DES, triple DES cryptographic classes for secret key digital signatures 
 *   <LI> PKI support: PKCS # 11(Netscape&#174), CSP (Microsoft&#174), X509 certificates 
 *   <LI> Support for Win2000 log-on 
 *   <LI> Physically secure iButtonTM case zeroizes contents on tampering 
 *   <LI> 3V to 5V operating range 
 *   <LI> ESD protection >25,000V   
 *   <LI> Over 10 years of data retention
 * </UL>
 *
 * <H3> Usage </H3> 
 * 
 * <DL> 
 * <DD> See the usage examples in 
 * {@link com.dalsemi.onewire.container.CommandAPDU CommandAPDU}
 * to create <code>CommandAPDU</code>.
 * <DD> See the usage examples in 
 * {@link com.dalsemi.onewire.container.ResponseAPDU ResponseAPDU}
 * to create <code>ResponseAPDU</code>.
 * </DL>
 *
 * <H3> DataSheet </H3> 
 * <DL>
 * <DD><A HREF="http://www.ibutton.com/ibuttons/java.html"> 
 *              http://www.ibutton.com/ibuttons/java.html</A>
 * </DL>
 * 
 * @see com.dalsemi.onewire.container.CommandAPDU
 * @see com.dalsemi.onewire.container.ResponseAPDU
 *
 * @version   0.00, 28 Aug 2000
 * @author    YL
 */
public class OneWireContainer16
   extends OneWireContainer
{

   /** CLA byte in the <code>CommandAPDU</code> header. */
   static final byte CLA = ( byte ) 0xD0;

   /** INS byte in the <code>CommandAPDU</code> header. */
   static final byte INS = ( byte ) 0x95;

   /** size of password length in byte */
   public static final int PASSWORD_LENGTH_SIZE = 1;

   /** maximum length of password in byte */
   public static final int PASSWORD_SIZE = 8;

   /** size of AID length in byte */
   public static final int AID_LENGTH_SIZE = 1;

   /** maximum length of AID in byte */
   public static final int AID_SIZE = 16;

   /** offset of AID length in applet APDU data stream */
   public static final int AID_LENGTH_OFFSET = PASSWORD_LENGTH_SIZE
                                               + PASSWORD_SIZE;

   /** offset of AID name in applet APDU data stream */
   public static final int AID_NAME_OFFSET = PASSWORD_LENGTH_SIZE
                                             + PASSWORD_SIZE
                                             + AID_LENGTH_SIZE;

   /** size of applet file header in byte */
   public static final int APPLET_FILE_HEADER_SIZE = PASSWORD_LENGTH_SIZE
                                                        + PASSWORD_SIZE
                                                        + AID_LENGTH_SIZE
                                                        + AID_SIZE;

   /** default APDU data packet length in byte */
   public static int APDU_PACKET_LENGTH = 64;

   /** password */
   private String password;

   /** current CommandaAPDU sent */
   private CommandAPDU capdu = null;

   /** current <code>ResponseAPDU</code> received */
   private ResponseAPDU rapdu = null;

   /** JibComm object for communicating with adapter */
   private JibComm jibComm;

   /** default JibComm run time */
   private int runTime = 0;

   /**
    * Creates an empty OneWireContainer16.  Must call <code>setupContainer()</code>
    *  method before using this new container.<p>
    *
    * This is one of the methods to construct a OneWireContainer16.  The
    * others are through creating a OneWireContainer16 with parameters.
    *
    * @see #OneWireContainer16(DSPortAdapter,byte[])
    * @see #OneWireContainer16(DSPortAdapter,long)
    * @see #OneWireContainer16(DSPortAdapter,String)
    * @see #setupContainer(DSPortAdapter,byte[])
    * @see #setupContainer(DSPortAdapter,long)
    * @see #setupContainer(DSPortAdapter,String)
    */
   public OneWireContainer16 ()
   {
      super();
   }   // OneWireContainer16()

   /**
    * Creates a OneWireContainer16 with the provided adapter object
    * and the address of this Java iButton.
    *
    * This is one of the methods to construct a OneWireContainer16.  The
    * others are through creating a OneWireContainer16 with different parameters
    * types.
    *
    * @param  sourceAdapter     adapter object required to communicate with
    *                           this Java iButton
    * @param  newAddress        address of this Java iButton
    *
    * @see com.dalsemi.onewire.utils.Address    
    * @see #OneWireContainer16()
    * @see #OneWireContainer16(DSPortAdapter,long)
    * @see #OneWireContainer16(DSPortAdapter,String)    
    */
   public OneWireContainer16 (DSPortAdapter sourceAdapter, byte[] newAddress)
   {
      super(sourceAdapter, newAddress);
   }

   /**
    * Creates a OneWireContainer16 with the provided adapter object
    * and the address of this Java iButton.
    *
    * This is one of the methods to construct a OneWireContainer16.  The
    * others are through creating a OneWireContainer16 with different parameters
    * types.
    *
    * @param  sourceAdapter     adapter object required to communicate with
    *                           this Java iButton
    * @param  newAddress        address of this Java iButton
    *
    * @see com.dalsemi.onewire.utils.Address    
    * @see #OneWireContainer16()
    * @see #OneWireContainer16(DSPortAdapter,byte[])
    * @see #OneWireContainer16(DSPortAdapter,String)    
    */
   public OneWireContainer16 (DSPortAdapter sourceAdapter, long newAddress)
   {
      super(sourceAdapter, newAddress);
   }

   /**
    * Creates a OneWireContainer16 with the provided adapter object
    * and the address of this Java iButton.
    *
    * This is one of the methods to construct a OneWireContainer16.  The
    * others are through creating a OneWireContainer16 with different parameters
    * types.
    *
    * @param  sourceAdapter     adapter object required to communicate with
    *                           this Java iButton
    * @param  newAddress        address of this Java iButton
    *
    * @see com.dalsemi.onewire.utils.Address    
    * @see #OneWireContainer16()
    * @see #OneWireContainer16(DSPortAdapter,byte[])
    * @see #OneWireContainer16(DSPortAdapter,long)
    */
   public OneWireContainer16 (DSPortAdapter sourceAdapter, String newAddress)
   {
      super(sourceAdapter, newAddress);
   }


   //-------------------------------------------------------------------------
   //-------- Methods
   //-------------------------------------------------------------------------
   //-------------------------------------------------------------------------

   /** Gets the Dallas Semiconductor part number of this Java iButton
    *  as a string.  For example "DS195X".
    *
    *  @return string represetation of this Java iButton name
    */
   public String getName ()
   {
        return "DS195X";      
   }   // getName()

   /** Gets the alternate Dallas Semiconductor part numbers or names.
    *  A 'family' of One-Wire devices may have more than one part number
    *  depending on packaging.  There can also be nicknames such as
    *  'Crypto iButton'.
    *
    *  @return string represetation of the alternate names
    */
   public String getAlternateNames ()
   {
      return "Java iButton, Cryptographic iButton";
   }   // getAlternateNames

   /** Gets a short description of the function of this Java iButton.
    *
    *  @return string represetation of the function description
    */
   public String getDescription ()
   {
      return "JavaCard 2.0 compliant device.";
   }   // getDescription()


   /**
    * Provides an adapter object to access this Java iButton.
    *
    * @param  sourceAdapter     adapter object required to communicate with
    *                           this Java iButton
    * @param  newAddress        address of this One-Wire device as a byte array.
    *
    * @see Address    
    */
   public void setupContainer (DSPortAdapter sourceAdapter, byte[] newAddress)
   {
      super.setupContainer(sourceAdapter, newAddress);
      setupJibComm(adapter, address);
   }

   /**
    * Provides an adapter object to access this Java iButton.
    *
    * @param  sourceAdapter     adapter object required to communicate with
    *                           this Java iButton
    * @param  newAddress        address of this One-Wire device as a long
    *
    * @see Address    
    */
   public void setupContainer (DSPortAdapter sourceAdapter, long newAddress)
   {
      super.setupContainer(sourceAdapter, newAddress);
      setupJibComm(adapter, address);
   }

   /**
    * Provides an adapter object to access this Java iButton.
    *
    * @param  sourceAdapter     adapter object required to communicate with
    *                           this Java iButton
    * @param  newAddress        address of this One-Wire device as a string
    *
    * @see Address    
    */
   public void setupContainer (DSPortAdapter sourceAdapter, String newAddress)
   {
      super.setupContainer(sourceAdapter, newAddress);
      setupJibComm(adapter, address);
   }


   /**
    * Provides an <code>JibComm</code> object to communicate 
    * with this Java iButton.
    *
    * @param  sourceAdapter     adapter object required to communicate with
    *                           this Java iButton
    * @param  newAddress        address of this One-Wire device as a byte array
    *
    * @see Address    
    */
   public void setupJibComm (DSPortAdapter sourceAdapter, byte[] newAddress)
   {
      // JibComm has to communicate at OVERDRIVE speed.
      speed           = adapter.SPEED_OVERDRIVE;
      speedFallBackOK = false;
      jibComm         = new JibComm(sourceAdapter, newAddress);
   }

   /**
    * Gets the maximum speed this One-Wire device can communicate at.
    *
    * @return maximum speed of this One-Wire device
    */
   public int getMaxSpeed ()
   {
      return adapter.SPEED_OVERDRIVE;
   }

   /**
   * Gets the current <code>CommandAPDU</code> sent to this Java iButton
   *
   * @return current <code>CommandAPDU</code> sent to this Java iButton
   */
   public CommandAPDU getCommandAPDUInfo ()
   {
      return capdu;
   }

   /**
   * Gets the current <code>ResponseAPDU</code> received from this Java iButton
   *
   * @return current <code>ResponseAPDU</code> received from this Java iButton
   */
   public ResponseAPDU getResponseAPDUInfo ()
   {
      return rapdu;
   }

   /**
   * Gets the run time value set for this Java iButton
   *
   * @return run time value of this Java iButton
   *
   * @see    #setRunTime
   */
   public int getRunTime ()
   {
      return runTime;
   }

   /**
    * Sets the run time value for this Java iButton
    *
    * @param  runTime a <code>4</code> bit value (<code>0 -15</code>) that 
    *                 represents the expected run time of the device
    *
    *  Actual run time is calculated as followed:
    *  <code><pre>
    *    runTime * 250 + 62.5 mS <BR>
    *  Therefore, 0 -> 0 * 250 + 62.5 =  62.5 mS
    *             1 -> 1 * 250 + 62.5 = 312.5 mS </pre></code>
    *  and so on.
    *
    * @throws IllegalArgumentException Invalid run time value
    *
    * @see #getRunTime        
    */
   public void setRunTime (int newRunTime)
      throws IllegalArgumentException
   {
      if ((runTime > 15) || (runTime < 0))
        throw new IllegalArgumentException("Run Time value should be between 0 and 15.");

      runTime = newRunTime;
   }

   /**
    * Sets the size of the packets sent on load.
    * The range can be from <code>64</code> to <code>112</code> bytes.
    *
    * @param  size new packet size
    *
    * @return <code>true</code> if packet size is set to the new value <BR>
    *         <code>false</code> if invalid packet size is requested
    *
    * @see #getLoadPacketSize
    * @see #loadApplet
    */
   public boolean setLoadPacketSize (int size)
   {

      // maximum number of bytes send is 124,
      // 4 for CommandAPDU header, 1 for data length,
      // 3 to append at sendAPDU(),  4 for JibComm setData header
      // and 112 for packet size
      if (size < 64 || size > 112)
         return false;

      APDU_PACKET_LENGTH = size;

      return true;
   }   // setLoadPacketSize()

   /**
    * Gets the size of the packets sent on load.
    *
    * @return  size of the packets sent on load
    *
    * @see #setLoadPacketSize
    * @see #loadApplet
    *    
    */
   public int getLoadPacketSize ()
   {
      return APDU_PACKET_LENGTH;
   }   // getLoadPacketSize()

   /**
    * Sets the PIN used to communicate with this Java iButton.
    * Once this method has been called, the PIN will be sent to every method
    * that requires a PIN.
    *
    * @param  passwd PIN to be set and sent to this Java iButton for each command
    *         that requires a PIN
    *
    * @see    #getCommandPINMode
    * @see    #setCommandPINMode
    * @see    #setCommonPIN    
    */
   public void setPIN (String passwd)
   {
      password = passwd;
   }   // setPIN()

   //-------------------------------------------------------------------------
   //-------- Firmware Command methods
   //-------------------------------------------------------------------------

   /**
    * Gets the amount of free RAM in this Java iButton.  No PIN required.
    *
    * @return <code>ResponseAPDU</code> with a data field containing the amount 
    *         of free RAM in this Java iButton.  Value returned is in little endian 
    *         format.
    *         <ul>
    *         <li>Success SW 0x9000.
    *         <li>For additional error codes, please see: 
    *             <A HREF="http://www.ibutton.com/jibkit/documents/sw.html">
    *                      http://www.ibutton.com/jibkit/documents/sw.html</A>
    *         </ul>
    *
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    *
    */
   public ResponseAPDU getFreeRAM ()
      throws OneWireException, OneWireIOException, IllegalArgumentException
   {
      capdu = new CommandAPDU(CLA, INS, ( byte ) 0x01, ( byte ) 0x01);
      rapdu = sendAPDU(capdu, runTime);

      return rapdu;
   }   // getFreeRAM()

   /**
    * Gets requested number of bytes of random data generated by this Java 
    * iButton. No PIN required.
    *
    * @param  numBytes the number of bytes requested.  <code>numByes</code> 
    *         should not exceed <code>119</code>.  If the number is greater 
    *         than <code>119</code>, the API only returns <code>119</code>.
    *
    * @return <code>ResponseAPDU</code> with a data field containing random 
    *         bytes generated
    *         <ul>
    *         <li>Success SW 0x9000.
    *         <li>For additional error codes, please see: 
    *             <A HREF="http://www.ibutton.com/jibkit/documents/sw.html">
    *                      http://www.ibutton.com/jibkit/documents/sw.html</A>
    *         </ul>
    *
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    */
   public ResponseAPDU getRandomBytes (int numBytes)
      throws OneWireException, OneWireIOException, IllegalArgumentException
   {

      // maximum number of bytes return is 124,
      // 2 for status word, 3 to discard and 119 for data
      if (numBytes > 119)
         numBytes = 119;

      byte[] data = new byte [2];

      data [0] = ( byte ) (numBytes & 0xFF);
      data [1] = ( byte ) ((numBytes >> 8) & 0xFF);
      capdu    = new CommandAPDU(CLA, INS, ( byte ) 0x01, ( byte ) 0x0D,
                                 data);
      rapdu    = sendAPDU(capdu, runTime);

      return rapdu;
   }   // getRandomBytes()

   /**
    * Gets the firmware version string. No PIN required.
    *
    * @return <code>ResponseAPDU</code> with a data field containing the firmware
    *         version string of this Java iButton
    *         <ul>
    *         <li>Success SW 0x9000.
    *         <li>For additional error codes, please see: 
    *             <A HREF="http://www.ibutton.com/jibkit/documents/sw.html">
    *                      http://www.ibutton.com/jibkit/documents/sw.html</A>
    *         </ul>
    *
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    */
   public ResponseAPDU getFirmwareVersionString ()
      throws OneWireException, OneWireIOException, IllegalArgumentException
   {
      capdu = new CommandAPDU(CLA, INS, ( byte ) 0x01, ( byte ) 0x00);
      rapdu = sendAPDU(capdu, runTime);

      return rapdu;
   }   // getFirmwareVersionString()

   /**
    * Gets the last error value.
    * If the error reporting mode is set (<code>1</code>), then this method 
    * will return the value of the last exception. No PIN required.
    *
    * @return <code>ResponseAPDU</code> containing the last exception value
    *         <ul>
    *         <li>Success SW 0x9000.
    *         <li>For additional error codes, please see: 
    *             <A HREF="http://www.ibutton.com/jibkit/documents/sw.html">
    *                      http://www.ibutton.com/jibkit/documents/sw.html</A>
    *         </ul>
    *
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    *
    * @see    #getErrorReportingMode
    * @see    #setErrorReportingMode    
    */
   public ResponseAPDU getLastError ()
      throws OneWireException, OneWireIOException, IllegalArgumentException
   {
      capdu = new CommandAPDU(CLA, INS, ( byte ) 0x01, ( byte ) 0x0F);
      rapdu = sendAPDU(capdu, runTime);

      return rapdu;
   }   // getLastError()

   /**
    * Gets the Error Reporting Mode.  This function is not supported in
    * the <code>.033</code> version of the firmware. No PIN required.
    *
    * @return <code>ResponseAPDU</code> containing the error reporting mode
    *         <ul>
    *         <li>Success SW 0x9000.
    *         <li>Failure SW 0x6A02. Command not supported by this Java iButton.
    *         <li>For additional error codes, please see: 
    *             <A HREF="http://www.ibutton.com/jibkit/documents/sw.html">
    *                      http://www.ibutton.com/jibkit/documents/sw.html</A>
    *         </ul>
    *
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    *
    * @see    #getLastError
    * @see    #setErrorReportingMode    
    */
   public ResponseAPDU getErrorReportingMode ()
      throws OneWireException, OneWireIOException, IllegalArgumentException
   {
      capdu = new CommandAPDU(CLA, INS, ( byte ) 0x01, ( byte ) 0x09);
      rapdu = sendAPDU(capdu, runTime);

      return rapdu;
   }   // getErrorReportingMode()

   /**
    * Sets the Error Reporting mode.
    * If the error reporting mode is set (<code>1</code>), this Java iButton
    * stores the last exception value code.  This code can be retreived by 
    * calling <code>getLastError()</code>.
    *
    *
    * @param mode <code>1</code> to turn on  Error Reporting Mode.
    *             <code>0</code> to turn off.
    * @return <code>ResponseAPDU</code> indicating success or failure
    *         <ul>
    *         <li>Success SW 0x9000.
    *         <li>Failure SW 0x6681. Invalid PIN.
    *         <li>For additional error codes, please see: 
    *             <A HREF="http://www.ibutton.com/jibkit/documents/sw.html">
    *                      http://www.ibutton.com/jibkit/documents/sw.html</A>
    *         </ul>
    *
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    *
    * @see    #getLastError
    * @see    #getErrorReportingMode
    */
   public ResponseAPDU setErrorReportingMode (int mode)
      throws OneWireException, OneWireIOException, IllegalArgumentException
   {
      byte[] data;

      if (password != null)
      {
         data     = new byte [2 + password.length()];
         data [0] = ( byte ) password.length();

         System.arraycopy(password.getBytes(), 0, data, 1, password.length());

         data [password.length() + 1] = ( byte ) mode;
      }
      else
      {
         data     = new byte [2];
         data [0] = ( byte ) 0;
         data [1] = ( byte ) mode;
      }

      capdu = new CommandAPDU(CLA, INS, ( byte ) 0x00, ( byte ) 0x09, data);
      rapdu = sendAPDU(capdu, runTime);

      return rapdu;
   }   // setErrorReportingMode()

   /**
    * Gets the AID of the applet by its installed number. No PIN required.
    *
    * @param index installed number of the applet on this Java iButton to get
    *              the AID of
    *
    * @return <code>ResponseAPDU</code> containing the AID of the applet 
    *         requested
    *         <ul>
    *         <li>Success SW 0x9000.
    *         <li>Failure SW 0x8453.  Applet not found.
    *             Valid applet index are in the range of <code>0-15</code>.
    *         <li>For additional error codes, please see: 
    *             <A HREF="http://www.ibutton.com/jibkit/documents/sw.html">
    *                      http://www.ibutton.com/jibkit/documents/sw.html</A>
    *         </ul>
    *
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    */
   public ResponseAPDU getAIDByNumber (int index)
      throws OneWireException, OneWireIOException, IllegalArgumentException
   {
      byte[] data = new byte [1];

      data [0] = ( byte ) (index & 0xFF);
      capdu    = new CommandAPDU(CLA, INS, ( byte ) 0x01, ( byte ) 0x0E,
                                 data);
      rapdu    = sendAPDU(capdu, runTime);

      if (rapdu.getSW() == 0x8453)   // applet not found
         return rapdu;

      // bug fix for Firmware 0.33
      int i = rapdu.getData().length;

      data = new byte [(i + 4)];

      System.arraycopy(rapdu.getData(), 0, data, 0, i);

      // return a "SUCCESS" ResponseAPDU
      data [i]     = rapdu.getSW1();
      data [i + 1] = rapdu.getSW2();
      data [i + 2] = ( byte ) 0x90;
      data [i + 3] = ( byte ) 0x00;

      return (new ResponseAPDU(data));
   }   // getAIDByNumber()

   /**
    * Deletes the currently selected applet.
    * If PIN protection is enabled, a PIN must be supplied.
    *
    * @return <code>ResponseAPDU</code> indicating success or failure
    *         <ul>
    *         <li>Success SW 0x9000.
    *         <li>Failure SW 0x6681. Invalid PIN.
    *         <li>For additional error codes, please see: 
    *             <A HREF="http://www.ibutton.com/jibkit/documents/sw.html">
    *                      http://www.ibutton.com/jibkit/documents/sw.html</A>
    *         </ul>
    *
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    *
    * @see    #deleteAppletByAID
    * @see    #deleteAppletByNumber
    */
   public ResponseAPDU deleteSelectedApplet ()
      throws OneWireException, OneWireIOException, IllegalArgumentException
   {
      byte[] data;

      if (password != null)
      {
         data     = new byte [1 + password.length()];
         data [0] = ( byte ) password.length();

         System.arraycopy(password.getBytes(), 0, data, 1, password.length());
      }
      else
      {
         data     = new byte [2];
         data [0] = ( byte ) 0;
      }

      capdu = new CommandAPDU(CLA, INS, ( byte ) 0x03, ( byte ) 0x00, data);
      rapdu = sendAPDU(capdu, runTime);

      return rapdu;
   }   // deleteSelectedApplet()

   /**
    * Deletes an applet by its installed number.
    *
    * @param  index installed number of the applet to delete
    *
    * @return <code>ResponseAPDU</code> indicating success or failure
    *         <ul>
    *         <li>Success SW 0x9000.
    *         <li>Failure SW 0x6681. Invalid PIN.
    *         <li>Failure SW 0x8453. Applet not found.
    *         <li>For additional error codes, please see: 
    *             <A HREF="http://www.ibutton.com/jibkit/documents/sw.html">
    *                      http://www.ibutton.com/jibkit/documents/sw.html</A>
    *         </ul>
    *
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    *
    * @see    #deleteAppletByAID
    * @see    #deleteSelectedApplet
    */
   public ResponseAPDU deleteAppletByNumber (int index)
      throws OneWireException, OneWireIOException, IllegalArgumentException
   {
      byte[] data;

      if (password != null)
      {
         data     = new byte [2 + password.length()];
         data [0] = ( byte ) password.length();

         System.arraycopy(password.getBytes(), 0, data, 1, password.length());

         data [password.length() + 1] = ( byte ) index;
      }
      else
      {
         data     = new byte [2];
         data [0] = ( byte ) 0;
         data [1] = ( byte ) index;
      }

      capdu = new CommandAPDU(CLA, INS, ( byte ) 0x03, ( byte ) 0x01, data);
      rapdu = sendAPDU(capdu, runTime);

      return rapdu;
   }   // deleleAppletByNumber()

   /**
    * Deletes an applet by its AID.
    *
    * @param  aid AID of applet to be deleted
    *
    * @return <code>ResponseAPDU</code> indicating success or failure
    *         <ul>
    *         <li>Success SW 0x9000.
    *         <li>Failure SW 0x6681. Invalid PIN.
    *         <li>Failure SW 0x8453. Applet not found.
    *         <li>For additional error codes, please see: 
    *             <A HREF="http://www.ibutton.com/jibkit/documents/sw.html">
    *                      http://www.ibutton.com/jibkit/documents/sw.html</A>
    *         </ul>
    *
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    *
    * @see    #deleteSelectedApplet
    * @see    #deleteAppletByNumber
    */
   public ResponseAPDU deleteAppletByAID (String aid)
      throws OneWireException, OneWireIOException, IllegalArgumentException
   {
      byte[] data;

      if (password != null)
      {
         data     = new byte [2 + password.length() + aid.length()];
         data [0] = ( byte ) password.length();

         System.arraycopy(password.getBytes(), 0, data, 1, password.length());

         data [password.length() + 1] = ( byte ) aid.length();

         System.arraycopy(aid.getBytes(), 0, data, password.length() + 2,
                          aid.length());
      }
      else
      {
         data     = new byte [2 + aid.length()];
         data [0] = ( byte ) 0;
         data [1] = ( byte ) aid.length();

         System.arraycopy(aid.getBytes(), 0, data, 2, aid.length());
      }

      capdu = new CommandAPDU(CLA, INS, ( byte ) 0x03, ( byte ) 0x02, data);
      rapdu = sendAPDU(capdu, runTime);

      return rapdu;
   }   // deleteAppletByAID()

   /**
    * Loads an applet onto this Java iButton.
    * This method takes an applet filename, directory
    * and AID.  The AID length must NOT exceed <code>AID_SIZE</code>.
    *
    * @param  fileName file name of the applet to be loaded into this Java iButton
    * @param  directoryName path to the applet to be loaded
    * @param  aid AID of the applet to be loaded
    *
    * @return <code>ResponseAPDU</code> indicating success or failure
    *         Possible return values are
    *         <ul>
    *         <li>Success SW 0x9000.
    *         <li>Failure SW 0x6400. Insufficient Memory
    *         <li>Failure SW 0x6901. Invalid AID Length
    *         <li>Failure SW 0x6902. Invalid API Version
    *         <li>Failure SW 0x6903. Invalid Password
    *         <li>Failure SW 0x6904. Invalid Signature Length
    *         <li>Failure SW 0x6905. Hash Corruption
    *         <li>Failure SW 0x6906. Hash Failure
    *         <li>Failure SW 0x6982. Invalid Signature
    *         <li>Failure SW 0x6A84. Class Length Overrun
    *         <li>Failure SW 0x6A86. Invalid Loader Command
    *         <li>Failure SW 0x6A87. Invalid Packet
    *         <li>For additional error codes, please see: 
    *             <A HREF="http://www.ibutton.com/jibkit/documents/sw.html">
    *                      http://www.ibutton.com/jibkit/documents/sw.html</A>
    *         </ul>
    *
    * @throws FileNotFoundException    Applet file not found
    * @throws IOException              Problem reading applet file
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    *
    * @see    #getLoadPINMode
    * @see    #setLoadPINMode    
    */
   public ResponseAPDU loadApplet (String fileName, String directoryName,
                                   String aid)
      throws OneWireException, OneWireIOException, IllegalArgumentException,
             FileNotFoundException, IOException
   {
      File            appletFile        = new File(directoryName, fileName);
      FileInputStream appletInputStream = new FileInputStream(appletFile);
      long            appletLength      = appletFile.length();
      long            blockSize         = appletInputStream.available();
      byte[]          appletBuffer      =
         new byte [APPLET_FILE_HEADER_SIZE + ( int ) appletLength];

      // Password FIX THIS TO HANDLE LOADING WITH A PASSWORD.
      if (password != null)
      {
         appletBuffer [0] = ( byte ) password.length();

         System.arraycopy(password.getBytes(), 0, appletBuffer, 1,
                          password.length());
      }

      // append AID to appletBuffer
      for (int offset = 0; offset < AID_SIZE; offset++)
      {
         if (offset < aid.length())
            appletBuffer [offset + AID_NAME_OFFSET] =
               ( byte ) aid.charAt(( int ) offset);
         else   // leave the rest in their default value (0)
            break;
      }

      appletBuffer [AID_LENGTH_OFFSET] = ( byte ) AID_SIZE;   // AID Length

      // get the applet file
      if ((appletLength) <= blockSize)   // in one chunk
         appletInputStream.read(appletBuffer, APPLET_FILE_HEADER_SIZE,
                                ( int ) appletLength);
      else
      {                                  // in multiple chunks
         long   bytesRead = 0;
         byte[] readBuffer;

         readBuffer = new byte [( int ) blockSize];

         while (bytesRead < appletLength)
         {
            appletInputStream.read(readBuffer);
            System.arraycopy(readBuffer, 0, appletBuffer,
                             ( int ) (bytesRead + APPLET_FILE_HEADER_SIZE),
                             readBuffer.length);

            bytesRead += blockSize;
         }
      }

      appletInputStream.close();

      int          i;
      int          bytesSent   = 0;
      boolean      firstPacket = true;
      ResponseAPDU rapdu       = null;
      byte[]       apduBuffer  = new byte [APDU_PACKET_LENGTH];

      while (bytesSent
             < (appletLength + APPLET_FILE_HEADER_SIZE - APDU_PACKET_LENGTH))
      {
         for (i = 0; (i < APDU_PACKET_LENGTH) && (i < appletBuffer.length);
                 i++)
            apduBuffer [i] = appletBuffer [i + bytesSent];

         bytesSent += APDU_PACKET_LENGTH;

         if (firstPacket == true)
         {
            capdu       = new CommandAPDU(CLA, ( byte ) 0xA6, ( byte ) 0x01,
                                          ( byte ) 0x00, apduBuffer);
            firstPacket = false;
         }
         else
         {
            capdu = new CommandAPDU(CLA, ( byte ) 0xA6, ( byte ) 0x02,
                                    ( byte ) 0x00, apduBuffer);
         }

         rapdu = sendAPDU(capdu, runTime);

         if (bytesSent < (appletLength + APPLET_FILE_HEADER_SIZE))
         {
            if (rapdu.getSW() != 0x6301)
               bytesSent = ( int ) (appletLength + APPLET_FILE_HEADER_SIZE
                                    + 1);
         }
      }

      if (bytesSent < (appletLength + APPLET_FILE_HEADER_SIZE))
      {
         apduBuffer =
            new byte [( int ) (appletLength + APPLET_FILE_HEADER_SIZE - bytesSent)];

         for (i = 0;
                 i < (appletLength + APPLET_FILE_HEADER_SIZE - bytesSent);
                 i++)
            apduBuffer [i] = appletBuffer [i + bytesSent];

         if (firstPacket == true)
         {
            capdu       = new CommandAPDU(CLA, ( byte ) 0xA6, ( byte ) 0x01,
                                          ( byte ) 0x00, apduBuffer);
            firstPacket = false;
         }
         else
         {
            capdu = new CommandAPDU(CLA, ( byte ) 0xA6, ( byte ) 0x02,
                                    ( byte ) 0x00, apduBuffer);
         }

         rapdu = sendAPDU(capdu, runTime);
      }

      return rapdu;
   }   // loadApplet()

   /**
    * Clears all memory in this Java iButton.  The erase will occur at the 
    * beginning of the next command.
    *
    * @return <code>ResponseAPDU</code> indicating success or failure
    *         <ul>
    *         <li>Success SW 0x9000.
    *         <li>Failure SW 0x6681. Invalid PIN.
    *         <li>For additional error codes, please see: 
    *             <A HREF="http://www.ibutton.com/jibkit/documents/sw.html">
    *                      http://www.ibutton.com/jibkit/documents/sw.html</A>    
    *         </ul>
    *
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    */
   public ResponseAPDU masterErase ()
      throws OneWireException, OneWireIOException, IllegalArgumentException
   {
      byte[] data;

      if (password != null)
      {
         data     = new byte [1 + password.length()];
         data [0] = ( byte ) password.length();

         System.arraycopy(password.getBytes(), 0, data, 1, password.length());
      }
      else
      {
         data     = new byte [1];
         data [0] = ( byte ) 0;
      }

      capdu = new CommandAPDU(CLA, INS, ( byte ) 0x00, ( byte ) 0x00, data);
      rapdu = sendAPDU(capdu, runTime);

      return rapdu;
   }   // masterErase()

   /**
    * Gets the number of times this Java iButton has been power-on-reset (POR)
    * since the last master erase.  Return value is in little endian
    * format. No PIN required.
    *
    * @return <code>ResponseAPDU</code> indicating success or failure
    *         <ul>
    *         <li>Success SW 0x9000. Data field contains the POR count.
    *         <li>For additional error codes, please see: 
    *             <A HREF="http://www.ibutton.com/jibkit/documents/sw.html">
    *                      http://www.ibutton.com/jibkit/documents/sw.html</A>
    *         </ul>
    *
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    *
    * @see    #masterErase
    */
   public ResponseAPDU getPORCount ()
      throws OneWireException, OneWireIOException, IllegalArgumentException
   {
      capdu = new CommandAPDU(CLA, INS, ( byte ) 0x02, ( byte ) 0x00);
      rapdu = sendAPDU(capdu, runTime);

      return rapdu;
   }   // getPORCount()

   /**
    * Gets the Real Time Clock.  No PIN required.
    *
    * @return <code>ResponseAPDU</code> containing the value of the real 
    *         time clock
    *         <ul>
    *         <li>Success SW 0x9000.
    *         <li>For additional error codes, please see: 
    *             <A HREF="http://www.ibutton.com/jibkit/documents/sw.html">
    *                      http://www.ibutton.com/jibkit/documents/sw.html</A>
    *         </ul>
    *
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    */
   public ResponseAPDU getRealTimeClock ()
      throws OneWireException, OneWireIOException, IllegalArgumentException
   {
      capdu = new CommandAPDU(CLA, INS, ( byte ) 0x01, ( byte ) 0x0C);
      rapdu = sendAPDU(capdu, runTime);

      return rapdu;
   }   // getRealTimeClock()

   /**
    * Gets the Answer To Reset (ATR) from this Java iButton. No PIN required.
    *
    * @return <code>ResponseAPDU</code> containing the ATR in the data field
    *         <ul>
    *         <li>Success SW 0x9000.
    *         <li>For additional error codes, please see: 
    *             <A HREF="http://www.ibutton.com/jibkit/documents/sw.html">
    *                      http://www.ibutton.com/jibkit/documents/sw.html</A>
    *         </ul>
    *
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    */
   public ResponseAPDU getATR ()
      throws OneWireException, OneWireIOException, IllegalArgumentException
   {
      capdu = new CommandAPDU(CLA, INS, ( byte ) 0x01, ( byte ) 0x0B);
      rapdu = sendAPDU(capdu, runTime);

      return rapdu;
   }   // getATR()

   /**
    * Gets the Ephemeral Gabage Collection Mode.
    * A value of <code>1</code> indicates ephemeral garbage collection 
    * is turned on, <code>0</code> if turned off. No PIN required.
    *
    * @return <code>ResponseAPDU</code> containing the Ephemeral Garbage 
    *         Collector Mode
    *         <ul>
    *         <li>Success SW 0x9000.
    *         <li>For additional error codes, please see: 
    *             <A HREF="http://www.ibutton.com/jibkit/documents/sw.html">
    *                      http://www.ibutton.com/jibkit/documents/sw.html</A>
    *         </ul>
    *
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    *
    * @see    #setEphemeralGCMode    
    */
   public ResponseAPDU getEphemeralGCMode ()
      throws OneWireException, OneWireIOException, IllegalArgumentException
   {
      capdu = new CommandAPDU(CLA, INS, ( byte ) 0x01, ( byte ) 0x02);
      rapdu = sendAPDU(capdu, runTime);

      return rapdu;
   }   // getEphemeralGCMode()

   /**
    * Gets the Applet Garbage Collection Mode.
    * A value of <code>1</code> indicates applet garbage collection 
    * is turned on, <code>0</code> if turned off. No PIN required.
    *
    * @return <code>ResponseAPDU</code> containing the Applet Garbage 
    *         Collector Mode
    *         <ul>
    *         <li>Success SW 0x9000.
    *         <li>For additional error codes, please see: 
    *             <A HREF="http://www.ibutton.com/jibkit/documents/sw.html">
    *                      http://www.ibutton.com/jibkit/documents/sw.html</A>
    *         </ul>
    *
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    *
    * @see    #setAppletGCMode    
    */
   public ResponseAPDU getAppletGCMode ()
      throws OneWireException, OneWireIOException, IllegalArgumentException
   {
      capdu = new CommandAPDU(CLA, INS, ( byte ) 0x01, ( byte ) 0x03);
      rapdu = sendAPDU(capdu, runTime);

      return rapdu;
   }   // getAppletGCMode()

   /**
    * Gets the Command PIN Mode.
    * A value of <code>1</code> indicates that a PIN is required for 
    * all Administrative and AID commands, <code>0</code> indicates 
    * free access. No PIN required.
    *
    * @return <code>ResponseAPDU</code> containing the Command PIN Mode
    *         <ul>
    *         <li>Success SW 0x9000.
    *         <li>For additional error codes, please see: 
    *             <A HREF="http://www.ibutton.com/jibkit/documents/sw.html">
    *                      http://www.ibutton.com/jibkit/documents/sw.html</A>
    *         </ul>
    *
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    *
    * @see    #setCommandPINMode
    * @see    #setCommonPIN
    * @see    #setPIN        
    */
   public ResponseAPDU getCommandPINMode ()
      throws OneWireException, OneWireIOException, IllegalArgumentException
   {
      capdu = new CommandAPDU(CLA, INS, ( byte ) 0x01, ( byte ) 0x04);
      rapdu = sendAPDU(capdu, runTime);

      return rapdu;
   }   // getCommandPINMode()

   /**
    * Gets the Load PIN Mode.
    * A value of <code>1</code> indicates that a PIN is required for Applet
    * loading. No PIN required.
    *
    * @return <code>ResponseAPDU</code> containing the Load PIN Mode
    *         <ul>
    *         <li>Success SW 0x9000.
    *         <li>For additional error codes, please see: 
    *             <A HREF="http://www.ibutton.com/jibkit/documents/sw.html">
    *                      http://www.ibutton.com/jibkit/documents/sw.html</A>
    *         </ul>
    *
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    *
    * @see    #loadApplet   
    * @see    #setLoadPINMode    
    */
   public ResponseAPDU getLoadPINMode ()
      throws OneWireException, OneWireIOException, IllegalArgumentException
   {
      capdu = new CommandAPDU(CLA, INS, ( byte ) 0x01, ( byte ) 0x05);
      rapdu = sendAPDU(capdu, runTime);

      return rapdu;
   }   // getLoadPINMode()

   /**
    * Gets the Restore Mode.
    * When Restore Mode is enabled (<code>1</code>), all field updates and
    * <code>javacard.framework.System</code> transactions are considered atomic.
    * If a tear occurs in the middle of these updates, values just
    * prior to the update are restored. No PIN required.
    *
    * @return <code>ResponseAPDU</code> containing the Restore Mode
    *         <ul>
    *         <li>Success SW 0x9000.
    *         <li>For additional error codes, please see: 
    *             <A HREF="http://www.ibutton.com/jibkit/documents/sw.html">
    *                      http://www.ibutton.com/jibkit/documents/sw.html</A>
    *         </ul>
    *
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    *
    * @see    #setRestoreMode        
    */
   public ResponseAPDU getRestoreMode ()
      throws OneWireException, OneWireIOException, IllegalArgumentException
   {
      capdu = new CommandAPDU(CLA, INS, ( byte ) 0x01, ( byte ) 0x06);
      rapdu = sendAPDU(capdu, runTime);

      return rapdu;
   }   // getRestoreMode()

   /**
    * Gets the Exception Mode.
    * When Exception Mode is enabled (<code>1</code>), Java API exceptions 
    * are thrown.  All uncaught exceptions return 0x6F00 in the SW.
    * When disabled, an error is returned from the VM. No PIN required.
    *
    * @return <code>ResponseAPDU</code> containing the Exception Mode value
    *         <ul>
    *         <li>Success SW 0x9000.
    *         <li>For additional error codes, please see: 
    *             <A HREF="http://www.ibutton.com/jibkit/documents/sw.html">
    *                      http://www.ibutton.com/jibkit/documents/sw.html</A>
    *         </ul>
    *
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    *
    * @see    #setExceptionMode    
    */
   public ResponseAPDU getExceptionMode ()
      throws OneWireException, OneWireIOException, IllegalArgumentException
   {
      capdu = new CommandAPDU(CLA, INS, ( byte ) 0x01, ( byte ) 0x07);
      rapdu = sendAPDU(capdu, runTime);

      return rapdu;
   }   // getExceptionMode()

   /**
    * Gets the size of the Commit Buffer.
    * Committing one field to the buffer requires <code>9</code> bytes. 
    * Therefore the default size of <code>72</code> bytes allows <code>8</code>
    * field updates. The minimum size allowed is <code>72</code> bytes and the 
    * maximum is restricted by the amount of free RAM. All values will be rounded
    * up to the next multiple of <code>9</code>. No PIN required.
    *
    * @return <code>ResponseAPDU</code> containing the Commit Buffer size in little
    *         endian format
    *         <ul>
    *         <li>Success SW 0x9000.
    *         <li>For additional error codes, please see: 
    *             <A HREF="http://www.ibutton.com/jibkit/documents/sw.html">
    *                      http://www.ibutton.com/jibkit/documents/sw.html</A>
    *         </ul>
    *
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    *
    * @see    #getFreeRAM    
    * @see    #setCommitBufferSize
    */
   public ResponseAPDU getCommitBufferSize ()
      throws OneWireException, OneWireIOException, IllegalArgumentException
   {
      capdu = new CommandAPDU(CLA, INS, ( byte ) 0x01, ( byte ) 0x0A);
      rapdu = sendAPDU(capdu, runTime);

      return rapdu;
   }   // getCommitBufferSize()

   /**
    * Sets the Common PIN.
    * This method is used to change the value of the Common PIN on the Java
    * iButton. If this Java iButton currently has a PIN, the <code>setPIN()</code>
    * method should be called prior to this method with the 'oldPIN' value.
    *
    * @param newPIN the value of the PIN to be set in this Java iButton
    *
    * @return <code>ResponseAPDU</code> indicating success or failure
    *         <ul>
    *         <li>Success SW 0x9000.
    *         <li>Failure SW 0x6681. Invalid PIN.
    *         <li>For additional error codes, please see: 
    *             <A HREF="http://www.ibutton.com/jibkit/documents/sw.html">
    *                      http://www.ibutton.com/jibkit/documents/sw.html</A>
    *         </ul>
    *
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    *
    * @see    #getCommandPINMode
    * @see    #setCommandPINMode
    * @see    #setPIN        
    */
   public ResponseAPDU setCommonPIN (String newPIN)
      throws OneWireException, OneWireIOException, IllegalArgumentException
   {
      byte[] data;

      if (password != null)
      {
         data     = new byte [newPIN.length() + password.length() + 2];
         data [0] = ( byte ) password.length();

         System.arraycopy(password.getBytes(), 0, data, 1, password.length());

         data [password.length() + 1] = ( byte ) newPIN.length();

         System.arraycopy(newPIN.getBytes(), 0, data, password.length() + 2,
                          newPIN.length());
      }
      else
      {
         data     = new byte [2 + newPIN.length()];
         data [0] = ( byte ) 0;
         data [1] = ( byte ) newPIN.length();

         System.arraycopy(newPIN.getBytes(), 0, data, 2, newPIN.length());
      }

      capdu = new CommandAPDU(CLA, INS, ( byte ) 0x00, ( byte ) 0x01, data);
      rapdu = sendAPDU(capdu, runTime);

      return rapdu;
   }   // setCommonPin()

   /**
    * Sets the Ephemeral Garbage Collection Mode.
    * A value of <code>1</code> turns the Ephemeral Garbage Collection on
    * and a value of <code>0</code> turns it off.
    *
    * @param mode <code>1</code> to turn on  Ephemeral Garbage Collection Mode.
    *             <code>0</code> to turn off.
    *
    * @return <code>ResponseAPDU</code> indicating success or failure
    *         <ul>
    *         <li>Success SW 0x9000.
    *         <li>Failure SW 0x6681. Invalid PIN.
    *         <li>For additional error codes, please see: 
    *             <A HREF="http://www.ibutton.com/jibkit/documents/sw.html">
    *                      http://www.ibutton.com/jibkit/documents/sw.html</A>
    *         </ul>
    *
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    *
    * @see    #getEphemeralGCMode
    */
   public ResponseAPDU setEphemeralGCMode (int mode)
      throws OneWireException, OneWireIOException, IllegalArgumentException
   {
      byte[] data;

      if (password != null)
      {
         data     = new byte [2 + password.length()];
         data [0] = ( byte ) password.length();

         System.arraycopy(password.getBytes(), 0, data, 1, password.length());

         data [password.length() + 1] = ( byte ) mode;
      }
      else
      {
         data     = new byte [2];
         data [0] = ( byte ) 0;
         data [1] = ( byte ) mode;
      }

      capdu = new CommandAPDU(CLA, INS, ( byte ) 0x00, ( byte ) 0x02, data);
      rapdu = sendAPDU(capdu, runTime);

      return rapdu;
   }   // set EphemeralGCMode()

   /**
    * Sets the Applet Garbage Collection Mode.
    * A value of <code>1</code> turns the Applet Garbage Collection on
    * and a value of <code>0</code> turns it off.
    *
    * @param mode <code>1</code> to turn on  Applet Garbage Collection Mode.
    *             <code>0</code> to turn off.
    *
    * @return <code>ResponseAPDU</code> indicating success or failure
    *         <ul>
    *         <li>Success SW 0x9000.
    *         <li>Failure SW 0x6681. Invalid PIN.
    *         <li>For additional error codes, please see: 
    *             <A HREF="http://www.ibutton.com/jibkit/documents/sw.html">
    *                      http://www.ibutton.com/jibkit/documents/sw.html</A>
    *         </ul>
    *
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    *
    * @see    #getAppletGCMode
    */
   public ResponseAPDU setAppletGCMode (int mode)
      throws OneWireException, OneWireIOException, IllegalArgumentException
   {
      byte[] data;

      if (password != null)
      {
         data     = new byte [2 + password.length()];
         data [0] = ( byte ) password.length();

         System.arraycopy(password.getBytes(), 0, data, 1, password.length());

         data [password.length() + 1] = ( byte ) mode;
      }
      else
      {
         data     = new byte [2];
         data [0] = ( byte ) 0;
         data [1] = ( byte ) mode;
      }

      capdu = new CommandAPDU(CLA, INS, ( byte ) 0x00, ( byte ) 0x03, data);
      rapdu = sendAPDU(capdu, runTime);

      return rapdu;
   }   // setAppletGCMode()

   /**
    * Sets the Command PIN Mode.
    * If command PIN mode is set to <code>1</code>, a PIN is required for all
    * Administrative commands.
    *
    * @param mode <code>1</code> to turn on  Command PIN Mode.
    *             <code>0</code> to turn off.
    *
    * @return <code>ResponseAPDU</code> indicating success or failure
    *         <ul>
    *         <li>Success SW 0x9000.
    *         <li>Failure SW 0x6681. Invalid PIN.
    *         <li>For additional error codes, please see: 
    *             <A HREF="http://www.ibutton.com/jibkit/documents/sw.html">
    *                      http://www.ibutton.com/jibkit/documents/sw.html</A>
    *         </ul>
    *
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    *
    * @see    #getCommandPINMode
    * @see    #setCommonPIN
    * @see    #setPIN    
    */
   public ResponseAPDU setCommandPINMode (int mode)
      throws OneWireException, OneWireIOException, IllegalArgumentException
   {
      byte[] data;

      if (password != null)
      {
         data     = new byte [2 + password.length()];
         data [0] = ( byte ) password.length();

         System.arraycopy(password.getBytes(), 0, data, 1, password.length());

         data [password.length() + 1] = ( byte ) mode;
      }
      else
      {
         data     = new byte [2];
         data [0] = ( byte ) 0;
         data [1] = ( byte ) mode;
      }

      capdu = new CommandAPDU(CLA, INS, ( byte ) 0x00, ( byte ) 0x04, data);
      rapdu = sendAPDU(capdu, runTime);

      return rapdu;
   }   // setCommandPINMode()

   /**
    * Sets the Load PIN Mode.
    * If the load PIN mode is set (<code>1</code>), then PINs are required
    * on applet loading.
    *
    * @param mode <code>1</code> to turn on  Load PIN Mode.
    *             <code>0</code> to turn off.
    *
    * @return <code>ResponseAPDU</code> indicating success or failure
    *         <ul>
    *         <li>Success SW 0x9000.
    *         <li>Failure SW 0x6681. Invalid PIN.
    *         <li>For additional error codes, please see: 
    *             <A HREF="http://www.ibutton.com/jibkit/documents/sw.html">
    *                      http://www.ibutton.com/jibkit/documents/sw.html</A>
    *         </ul>
    *
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    *
    * @see    #getLoadPINMode
    * @see    #loadApplet    
    */
   public ResponseAPDU setLoadPINMode (int mode)
      throws OneWireException, OneWireIOException, IllegalArgumentException
   {
      byte[] data;

      if (password != null)
      {
         data     = new byte [2 + password.length()];
         data [0] = ( byte ) password.length();

         System.arraycopy(password.getBytes(), 0, data, 1, password.length());

         data [password.length() + 1] = ( byte ) mode;
      }
      else
      {
         data     = new byte [2];
         data [0] = ( byte ) 0;
         data [1] = ( byte ) mode;
      }

      capdu = new CommandAPDU(CLA, INS, ( byte ) 0x00, ( byte ) 0x05, data);
      rapdu = sendAPDU(capdu, runTime);

      return rapdu;
   }   // setLoadPINMode()

   /**
    * Sets the Restore Mode.
    * When Restore Mode is enabled (<code>1</code>), all field updates and
    * <code>javacard.framework.System</code> transactions are considered atomic.
    * If a tear occurs in the middle of these updates, values just prior to the 
    * update are restored.
    *
    * @param mode <code>1</code> to turn on  Restore Mode.
    *             <code>0</code> to turn off.
    *
    * @return <code>ResponseAPDU</code> indicating success or failure
    *         <ul>
    *         <li>Success SW 0x9000.
    *         <li>Failure SW 0x6681. Invalid PIN.
    *         <li>For additional error codes, please see: 
    *             <A HREF="http://www.ibutton.com/jibkit/documents/sw.html">
    *                      http://www.ibutton.com/jibkit/documents/sw.html</A>
    *         </ul>
    *
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    *
    * @see    #getRestoreMode    
    */
   public ResponseAPDU setRestoreMode (int mode)
      throws OneWireException, OneWireIOException, IllegalArgumentException
   {
      byte[] data;

      if (password != null)
      {
         data     = new byte [2 + password.length()];
         data [0] = ( byte ) password.length();

         System.arraycopy(password.getBytes(), 0, data, 1, password.length());

         data [password.length() + 1] = ( byte ) mode;
      }
      else
      {
         data     = new byte [2];
         data [0] = ( byte ) 0;
         data [1] = ( byte ) mode;
      }

      capdu = new CommandAPDU(CLA, INS, ( byte ) 0x00, ( byte ) 0x06, data);
      rapdu = sendAPDU(capdu, runTime);

      return rapdu;
   }   // setRestoreMode()

   /**
    * Sets the Exception Mode.
    * When Exception Mode is enabled (<code>1</code>), Java API exceptions 
    * are thrown. All uncaught exceptions return 0x6F00 in the SW. When 
    * disabled, an error is returned from the VM.
    *
    * @param mode <code>1</code> to turn on  Exception Mode.
    *             <code>0</code> to turn off.    
    *    
    * @return <code>ResponseAPDU</code> indicating success or failure
    *         <ul>
    *         <li>Success SW 0x9000.
    *         <li>Failure SW 0x6681. Invalid PIN.
    *         <li>For additional error codes, please see: 
    *             <A HREF="http://www.ibutton.com/jibkit/documents/sw.html">
    *                      http://www.ibutton.com/jibkit/documents/sw.html</A>
    *         </ul>
    *
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    *
    * @see    #getExceptionMode    
    */
   public ResponseAPDU setExceptionMode (int mode)
      throws OneWireException, OneWireIOException, IllegalArgumentException
   {
      byte[] data;

      if (password != null)
      {
         data     = new byte [2 + password.length()];
         data [0] = ( byte ) password.length();

         System.arraycopy(password.getBytes(), 0, data, 1, password.length());

         data [password.length() + 1] = ( byte ) mode;
      }
      else
      {
         data     = new byte [2];
         data [0] = ( byte ) 0;
         data [1] = ( byte ) mode;
      }

      capdu = new CommandAPDU(CLA, INS, ( byte ) 0x00, ( byte ) 0x07, data);
      rapdu = sendAPDU(capdu, runTime);

      return rapdu;
   }   // setExceptionMode()

   /**
    * Sets the Commit Buffer size.
    * Committing one field to the buffer requires <code>9</code> bytes. 
    * Therefore the default size of <code>72</code> bytes allows <code>8</code>
    * field updates. The minimum size allowed is <code>72</code> bytes and the 
    * maximum is restricted by the amount of free RAM. All values will be rounded
    * up to the next multiple of <code>9</code>.
    *
    * @param  size size of the desired commit buffer
    *
    * @return <code>ResponseAPDU</code> indicating success or failure
    *         <ul>
    *         <li>Success SW 0x9000.
    *         <li>Failure SW 0x6681. Invalid PIN.
    *         <li>For additional error codes, please see: 
    *             <A HREF="http://www.ibutton.com/jibkit/documents/sw.html">
    *                      http://www.ibutton.com/jibkit/documents/sw.html</A>
    *         </ul>
    *
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    *
    * @see    #getCommitBufferSize    
    */
   public ResponseAPDU setCommitBufferSize (int size)
      throws OneWireException, OneWireIOException, IllegalArgumentException
   {
      byte[] data;

      if (password != null)
      {
         data     = new byte [3 + password.length()];
         data [0] = ( byte ) password.length();

         System.arraycopy(password.getBytes(), 0, data, 1, password.length());

         data [password.length() + 1] = ( byte ) (size & 0xFF);
         data [password.length() + 2] = ( byte ) (size >>> 8);
      }
      else
      {
         data     = new byte [3];
         data [0] = ( byte ) 0;
         data [1] = ( byte ) (size & 0xFF);
         data [2] = ( byte ) (size >>> 8);
      }

      capdu = new CommandAPDU(CLA, INS, ( byte ) 0x00, ( byte ) 0x0A, data);
      rapdu = sendAPDU(capdu, runTime);

      return rapdu;
   }   // setCommitBufferSize()

   /**
    * Selects an applet by AID.
    * Lenght of AID should not exceed <code>AID_SIZE</code>. No PIN required.
    *
    * @param  aid the AID of the applet to be selected
    *
    * @return <code>ResponseAPDU</code> indicating success or failure
    *         <ul>
    *         <li>Success SW 0x9000.
    *         <li>Failure SW 0x6A82. Unable to Select Applet
    *         <li>Failure SW 0x8453. Applet not found.
    *         <li>For additional error codes, please see: 
    *             <A HREF="http://www.ibutton.com/jibkit/documents/sw.html">
    *                      http://www.ibutton.com/jibkit/documents/sw.html</A>
    *         </ul>
    *
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    *
    * @see    #deleteAppletByAID
    * @see    #deleteAppletByNumber    
    * @see    #deleteSelectedApplet
    * @see    #getAIDByNumber
    * @see    #process
    */
   public ResponseAPDU select (String aid)
      throws OneWireException, OneWireIOException, IllegalArgumentException
   {
      if (aid.length() > AID_SIZE)
         throw new IllegalArgumentException("AID length should not exceed "
                                            + AID_SIZE);

      byte[] data = new byte [AID_SIZE];

      System.arraycopy(aid.getBytes(), 0, data, 0, aid.length());

      capdu = new CommandAPDU(( byte ) 0x00, ( byte ) 0xA4, ( byte ) 0x04,
                              ( byte ) 0x00, data);
      rapdu = sendAPDU(capdu, runTime);

      return rapdu;
   }   // select

   /**
    * Sends a generic process command to this JavaiButton.
    * This method can be used to send any <code>CommandAPDU</code>.
    * No PIN required.
    *
    * @param capdu <code>CommandAPDU</code> to be sent to this Java iButton
    *
    * @return <code>ResponseAPDU</code> the response from this Java iButton
    *
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    *
    * @see    #select
    */
   public ResponseAPDU process (CommandAPDU capdu)
      throws OneWireException, OneWireIOException, IllegalArgumentException
   {
      this.capdu = capdu;
      rapdu      = sendAPDU(capdu, runTime);

      return rapdu;
   }

   /**
    * Sends a <code>CommandAPDU</code> to this Java iButton.
    * No PIN required.
    *
    * @param  capdu the <code>CommandAPDU</code> to be sent to this Java iButton
    *
    * @param  runTime a <code>4</code> bit value (<code>0 -15</code>) that 
    *                 represents the expected run time of the device
    *
    *  <code><pre>
    *    runTime * 250 + 62.5 mS <BR>
    *  Therefore, 0 -> 0 * 250 + 62.5 =  62.5 mS
    *             1 -> 1 * 250 + 62.5 = 312.5 mS </pre></code>
    *  and so on.
    *
    * @return <code>ResponseAPDU</code> the <code>ResponseAPDU</code> from
    *         this Java iButton
    *
    * @throws IllegalArgumentException Invalid run time value
    * @throws OneWireException         Part could not be found [ fatal ]
    * @throws OneWireIOException       Data wasn't transferred properly [ recoverable ]
    */
   public ResponseAPDU sendAPDU (CommandAPDU capdu, int runTime)
      throws OneWireException, OneWireIOException, IllegalArgumentException
   {

      if ((runTime > 15) || (runTime < 0))
        throw new IllegalArgumentException("Run Time value should be between 0 and 15.");

      int    capduLength = capdu.getLength();
      byte[] apdu        = new byte [capduLength + 3];

      // append extra bytes for legacy reasons
      apdu [0] = ( byte ) (2 + capduLength);
      apdu [1] = ( byte ) 0x89;
      apdu [2] = ( byte ) 0x00;

      System.arraycopy(capdu.getBytes(), 0, apdu, 3, capduLength);
      // Attempt to put adapter into OverDrive Mode.
      doSpeed();
      rapdu = new ResponseAPDU(jibComm.transferJibData(apdu, runTime));

      return rapdu;
   }
}
